#!/usr/bin/env node
import axios from 'axios';
import dedent from 'ts-dedent';
import create from 'zustand';
import { immer } from 'zustand/middleware/immer';
import { last } from 'radash';
import { useZustand } from 'use-zustand';
import { v4 } from 'uuid';
import fs from 'fs';
import path from 'path';
import { homedir } from 'os';
import { Command } from 'commander';
import chalk from 'chalk';
import open from 'open';
import ora from 'ora';
import { jsx, jsxs, Fragment } from 'react/jsx-runtime';
import { Box, Text, useApp, useInput, render } from 'ink';
import { useQuery, useMutation, useQueryClient, QueryClient, QueryClientProvider } from 'react-query';
import { highlight } from 'cli-highlight';
import getStream from 'get-stream';
import SelectInput from 'ink-select-input';
import { useEffect } from 'react';
import { marked } from 'marked';
import TerminalRenderer from 'marked-terminal';
import { UncontrolledTextInput } from 'ink-text-input';

const states = [
  /** Awaiting a completion from the model */
  "initial",
  /** A response has been received and is being displayed */
  "response",
  /** The user has accepted a completion and may be asked to confirm before
  running it */
  "confirm",
  /** The user is revising the query */
  "revise"
];
const makeInitialState = (initialInput, mode) => {
  return {
    mode,
    state: states[0],
    history: {
      iteration: 0,
      lastQuery: initialInput,
      interactions: []
    }
  };
};
const useCliStore = create(
  immer((set, get) => ({
    mode: "",
    id: v4(),
    state: states[0],
    history: {
      iteration: 0,
      lastQuery: "",
      interactions: []
    },
    errors: {
      completion: void 0
    },
    actions: {
      setExitReason: (exitReason) => {
        set((state) => {
          state.exitReason = exitReason;
        });
      },
      setState: (newState) => {
        set((state) => {
          state.state = newState;
        });
      },
      handleRevision: (revision) => {
        set((state) => {
          state.history.lastQuery = revision;
          state.state = "initial";
        });
      },
      handleResponse: (rawResponse) => {
        let response = rawResponse.type == "refuse" ? rawResponse : {
          type: "completion",
          completion: dedent(rawResponse.completion)
        };
        set((state) => {
          state.history.interactions.push({
            query: state.history.lastQuery,
            response
          });
          state.history.lastQuery = void 0;
          state.state = "response";
        });
      }
    }
  }))
);
const useActions = () => useZustand(useCliStore, (state) => state.actions);
const useHistory = () => useZustand(useCliStore, (state) => state.history);
const useState = () => useZustand(useCliStore, (state) => state.state);
const useLastCompletion = () => useZustand(useCliStore, (state) => {
  var _a;
  const completion = (_a = last(state.history.interactions)) == null ? void 0 : _a.response;
  if (!completion || completion.type !== "completion") {
    return void 0;
  }
  return completion.completion;
});

const iKeyStd = "8c0935a9-0e45-4f28-b91f-2813d3b5694a";
const connectionStringStd = "InstrumentationKey=8c0935a9-0e45-4f28-b91f-2813d3b5694a;IngestionEndpoint=https://westus2-2.in.applicationinsights.azure.com/;LiveEndpoint=https://westus2.livediagnostics.monitor.azure.com/";
const iKeySec = "0e2e2e95-4781-49f1-8f01-40ca86b28736";
const connectionStringSec = "InstrumentationKey=0e2e2e95-4781-49f1-8f01-40ca86b28736;IngestionEndpoint=https://eastus-8.in.applicationinsights.azure.com/;LiveEndpoint=https://eastus.livediagnostics.monitor.azure.com/";
const useTelemetryStore = create(
  immer(() => ({
    clients: {
      secure: void 0,
      standard: void 0
    }
  }))
);
function genericSender(client, name) {
  return (event, properties, measurements) => {
    if (client === void 0) {
      console.error(`Telemetry client ${name} not initialized`);
    }
    client == null ? void 0 : client.trackEvent({
      name: event,
      properties,
      measurements: { ...measurements }
    });
    client == null ? void 0 : client.flush();
  };
}
const useSecureSender = () => {
  const { clients } = useTelemetryStore.getState();
  const secure = clients.secure;
  if (!secure)
    throw new Error("Secure telemetry client not initialized");
  return genericSender(
    secure,
    "secure"
  );
};
const useStandardSender = () => {
  const { clients } = useTelemetryStore.getState();
  const standard = clients.standard;
  if (!standard)
    throw new Error("Standard telemetry client not initialized");
  return genericSender(
    standard,
    "standard"
  );
};
const telemetryScrubber = (envelope) => {
  envelope.tags["ai.cloud.roleInstance"] = "REDACTED";
  envelope.tags["ai.device.osVersion"] = "REDACTED";
  envelope.tags["ai.location.ip"] = "0.0.0.0";
  return true;
};
async function initializeTelemetry() {
  try {
    const appInsights = await import('applicationinsights');
    let secureClient = new appInsights.TelemetryClient(connectionStringSec);
    let standardClient = new appInsights.TelemetryClient(connectionStringStd);
    secureClient.addTelemetryProcessor(telemetryScrubber);
    standardClient.addTelemetryProcessor(telemetryScrubber);
    useTelemetryStore.setState({
      clients: {
        secure: secureClient,
        standard: standardClient
      }
    });
  } catch (e) {
    throw new Error("Failed to initialize telemetry client: " + (e == null ? void 0 : e.message));
  }
}
function flushTelemetry() {
  const { clients } = useTelemetryStore.getState();
  return new Promise((resolve) => {
    var _a, _b;
    (_a = clients.secure) == null ? void 0 : _a.flush();
    (_b = clients.standard) == null ? void 0 : _b.flush();
    resolve();
  });
}

function iteration(history) {
  return history.interactions.length + (history.lastQuery ? 1 : 0);
}
function isRefuseResponse(history) {
  var _a;
  return history.lastQuery === void 0 && ((_a = last(history.interactions)) == null ? void 0 : _a.response.type) === "refuse";
}
function lastResponse(history) {
  var _a;
  if (history.lastQuery !== void 0) {
    return void 0;
  }
  return (_a = last(history.interactions)) == null ? void 0 : _a.response;
}

const DEVICE_CODE_URL = "https://github.com/login/device/code";
const ACCESS_TOKEN_URL = "https://github.com/login/oauth/access_token";
const COPILOT_TOKEN_URL = "https://api.github.com/copilot_internal/v2/token";
const CLIENT_ID = "Iv1.b507a08c87ecfe98";
const ACCESS_TOKEN_FILENAME = ".copilot-cli-access-token";
const COPILOT_TOKEN_FILENAME = ".copilot-cli-copilot-token";
const NEXT_WAITLIST_URL = "https://next-waitlist.azurewebsites.net";
const WAITLIST_PROJECT_ID = "copilot-cli";

class CopilotToken {
  constructor(token) {
    const firstPart = token.split(":")[0];
    const fields = firstPart == null ? void 0 : firstPart.split(";");
    this.tokenMap = /* @__PURE__ */ new Map();
    for (const field of fields) {
      const [key, value] = field.split("=");
      this.tokenMap.set(key, value);
    }
  }
  getTokenValue(key) {
    return this.tokenMap.get(key);
  }
}
function getCopilotTelemetryConfig() {
  let token = getCachedCopilotToken();
  let parsedToken = new CopilotToken(token);
  return {
    tid: parsedToken.getTokenValue("tid"),
    rt: parsedToken.getTokenValue("rt")
  };
}
function getCachedAccessToken() {
  const filename = path.join(homedir(), ACCESS_TOKEN_FILENAME);
  const token = fs.readFileSync(filename, "utf8");
  return token;
}
function getCachedCopilotToken() {
  const filename = path.join(homedir(), COPILOT_TOKEN_FILENAME);
  const token = fs.readFileSync(filename, "utf8");
  return token;
}
function setCachedAccessToken(token) {
  const filename = path.join(homedir(), ACCESS_TOKEN_FILENAME);
  fs.writeFileSync(filename, token, {
    mode: 384
  });
}
function setCachedCopilotToken(token) {
  const filename = path.join(homedir(), COPILOT_TOKEN_FILENAME);
  fs.writeFileSync(filename, token, {
    mode: 384
  });
}
function tokenPresenceCheck() {
  let accessToken = getCachedAccessToken();
  let copilotToken = getCachedCopilotToken();
  if (!accessToken || !copilotToken) {
    throw new Error("Missing cached token");
  }
}

// Generated by genversion.
const version$1 = '0.1.16';

function telemetrizeInteractions(history) {
  function responseAsString(response) {
    switch (response.type) {
      case "completion":
        return response.completion;
      case "refuse":
        return "No completion: " + response.message;
    }
  }
  let s = history.interactions.map(
    (interaction) => `Q:
${interaction.query}
A:
${responseAsString(
      interaction.response
    )}
`
  ).join("---\n");
  if (history.lastQuery) {
    s += `Q:
${history.lastQuery}
`;
  }
  return s;
}
function makeApiPayload({
  key,
  event,
  properties,
  measurements
}) {
  return {
    name: "Microsoft.ApplicationInsights.Event",
    time: (/* @__PURE__ */ new Date()).toISOString(),
    iKey: key,
    tags: {},
    data: {
      baseType: "EventData",
      baseData: {
        ver: 2,
        name: event,
        properties,
        measurements
      }
    }
  };
}
const telemetry = {
  _buildPayload: (event, stdProps = {}, secProps = {}) => {
    const { id: sessionId, history, mode } = useCliStore.getState();
    const { rt: flagSecureTelemetry, tid: userTrackingId } = getCopilotTelemetryConfig();
    const commonProperties = {
      userTrackingId: userTrackingId != null ? userTrackingId : "",
      sessionId,
      mode,
      copilotCLIVersion: version$1,
      ...stdProps
      // These take precedence
    };
    const commonMeasurements = {
      iteration: iteration(history)
    };
    const secureProperties = {
      ...commonProperties,
      ...history.lastQuery && { lastQuery: history.lastQuery },
      interactions: telemetrizeInteractions(history),
      ...secProps
      // These take precendece
    };
    return {
      commonProperties,
      commonMeasurements,
      secureProperties,
      flagSecureTelemetry
    };
  },
  send: (event, stdProps = {}, secProps = {}) => {
    const payload = telemetry._buildPayload(event, stdProps, secProps);
    const senderStd = useStandardSender();
    const senderSec = useSecureSender();
    senderStd(event, payload.commonProperties, payload.commonMeasurements);
    if (payload.flagSecureTelemetry === "1") {
      senderSec(event, payload.secureProperties, payload.commonMeasurements);
    }
  },
  sendApi: (event, stdProps = {}, secProps = {}) => {
    const payload = telemetry._buildPayload(event, stdProps, secProps);
    let aiAxios = axios.create({
      baseURL: `https://dc.services.visualstudio.com/v2`
    });
    return Promise.all([
      aiAxios.post(
        `/track`,
        makeApiPayload({
          key: iKeyStd,
          event,
          properties: payload.commonProperties,
          measurements: payload.commonMeasurements
        })
      ),
      payload.flagSecureTelemetry === "1" && aiAxios.post(
        `/track`,
        makeApiPayload({
          key: iKeySec,
          event,
          properties: payload.secureProperties,
          measurements: payload.commonMeasurements
        })
      )
    ]);
  }
};
function trackInvocation() {
  telemetry.send("invocation");
}
function trackModelResponse() {
  const { history } = useCliStore.getState();
  const last = lastResponse(history);
  if (last === void 0) {
    console.log("TELEMETRY ERROR: No last response");
  } else {
    if (last.type === "completion") {
      telemetry.send("completion-success");
    } else {
      telemetry.send("completion-refuse");
    }
  }
}
function trackExplanationSuccess(explanation) {
  telemetry.send(
    "explanation-success",
    {},
    {
      explanation
    }
  );
}
function trackExplanationError(message) {
  telemetry.send("explanation-error", {}, { message });
}
function trackRevisionRequest() {
  telemetry.send("revision-request");
}
function trackConfirmCommandRun() {
  return telemetry.sendApi("confirm-command-run");
}
function trackConfirmCommandCancel() {
  telemetry.send("confirm-command-cancel");
}
function trackExit(exitReason) {
  return telemetry.sendApi("exit", { exitReason });
}

function indentString(string, count = 1, options = {}) {
	const {
		indent = ' ',
		includeEmptyLines = false
	} = options;

	if (typeof string !== 'string') {
		throw new TypeError(
			`Expected \`input\` to be a \`string\`, got \`${typeof string}\``
		);
	}

	if (typeof count !== 'number') {
		throw new TypeError(
			`Expected \`count\` to be a \`number\`, got \`${typeof count}\``
		);
	}

	if (count < 0) {
		throw new RangeError(
			`Expected \`count\` to be at least 0, got \`${count}\``
		);
	}

	if (typeof indent !== 'string') {
		throw new TypeError(
			`Expected \`options.indent\` to be a \`string\`, got \`${typeof indent}\``
		);
	}

	if (count === 0) {
		return string;
	}

	const regex = includeEmptyLines ? /^/gm : /^(?!\s*$)/gm;

	return string.replace(regex, indent.repeat(count));
}

const DONTKNOW = "@DONTKNOW@";
function assistHeadline(content) {
  return dedent`
    Suggest ${content} based on natural language queries.
    The user may revise their query, and an updated answer will be provided.
    Below 'Q' is followed by a question, 'A' by an answer, and 'R' follwed by a revision to the query.
  `;
}
function interact(query, completion) {
  return {
    query,
    response: {
      type: "completion",
      completion
    }
  };
}
const EXPLAIN_LIST_INDENT = "  ";
function explanation(query, explanation2) {
  function formatBullet(bullet) {
    return "* " + bullet;
  }
  function flattenSub(sub, depth) {
    return sub.map((s) => flatten(s, depth)).join("\n");
  }
  function flatten(e, depth) {
    if (typeof e === "string") {
      return EXPLAIN_LIST_INDENT.repeat(depth) + formatBullet(e);
    }
    const head = EXPLAIN_LIST_INDENT.repeat(depth) + formatBullet(e.hd);
    if (e.sub !== void 0 && e.sub.length > 0) {
      return head + "\n" + flattenSub(e.sub, depth + 1);
    } else {
      return head;
    }
  }
  return [interact(query, flattenSub(explanation2, 0))];
}
const Q_A_INDENT = 4;
function formatQA(query, answer, question_or_revision) {
  const questionMark = question_or_revision === "question" ? "Q" : "R";
  const questionPart = indentString(query, Q_A_INDENT);
  const answerPart = answer === void 0 || answer === "" ? "" : indentString(answer, Q_A_INDENT);
  return `${questionMark}:
${questionPart}
A:
${answerPart}`;
}
function responseAsAnswer(response) {
  return response.type === "completion" ? response.completion : DONTKNOW;
}
function formatInteractionsAsQA(interactions) {
  return interactions.map(
    (i, idx) => formatQA(
      i.query,
      responseAsAnswer(i.response),
      idx > 0 ? "revision" : "question"
    )
  ).join("\n");
}
function formatHistoryAsQ(history) {
  if (history.lastQuery === void 0) {
    return formatInteractionsAsQA(history.interactions);
  } else {
    const lastInteraction = {
      query: history.lastQuery,
      response: { type: "completion", completion: "" }
    };
    return formatInteractionsAsQA([...history.interactions, lastInteraction]);
  }
}
function fewShotPrompter(headline, examples, formatOpenQuestion) {
  return (query) => {
    const queryString = formatOpenQuestion(query);
    const body = [
      ...examples.map((e) => formatInteractionsAsQA(e)),
      queryString
    ].join("\n---\n");
    return `${headline}

${body}`;
  };
}
const fewShotQueryPrompter = (headline, examples) => fewShotPrompter(headline, examples, formatHistoryAsQ);
function answerFromCompletion(completion) {
  const lines = completion.split("\n").filter((line) => line !== "");
  const stripIndent = " ".repeat(Q_A_INDENT);
  if (lines.every((line) => line.startsWith(stripIndent))) {
    return lines.map((line) => line.slice(Q_A_INDENT)).join("\n");
  }
  return completion;
}

const shellPromptHeadline = assistHeadline("shell commands");
const shellFewshotExamples = [
  [interact("Switch to the main branch", "git checkout main")],
  [
    interact(
      "Find ts files. Filter out node_modules",
      `find . -name "*.ts" | grep -v node_modules`
    )
  ],
  [
    interact(
      'Move the latest commit to new branch "new-feature"',
      [
        "git branch new-feature",
        "git reset --hard HEAD~1",
        "git switch new-feature"
      ].join("\n")
    )
  ],
  [interact("Perfoogling the clubberflob", DONTKNOW)],
  [
    interact("Find all photos with a cat", DONTKNOW),
    interact("with cat in the name", `find . -name "*cat*"`),
    interact("png or jpg", `find . -name "*cat*" | grep -E ".(png|jpg)$"`)
  ],
  [
    interact(
      "Loop over all .py files, printing the file name and number of lines",
      [
        `for file in $(find -name "*.py"); do`,
        `  echo "$file";`,
        `  wc -l $file;`,
        `done`
      ].join("\n")
    )
  ],
  [interact("Whats my public IP address?", "curl ifconfig.me")],
  [
    interact(
      "For each folder, get the variable TITLE from Makefile",
      [
        `for folder in */; do`,
        `  echo $(grep TITLE $folder/Makefile | cut -d "=" -f 2);`,
        `done`
      ].join("\n")
    ),
    interact(
      "assign to the variable T",
      [
        `for folder in */; do`,
        `  T=$(grep TITLE $folder/Makefile | cut -d "=" -f 2);`,
        `  echo $T;`,
        `done`
      ].join("\n")
    ),
    interact(
      "copy the contents of template.md to build/<folder> replacing TITLE with T",
      [
        `for folder in */; do`,
        `  T=$(grep TITLE $folder/Makefile | cut -d "=" -f 2);`,
        `  cp template.md build/$folder.md;`,
        `  sed -i "s/TITLE/$T/g" build/$folder/README.md;`,
        `done`
      ].join("\n")
    )
  ],
  [
    interact("Install Typescript compiler", "sudo apt-get install typescript"),
    interact("on mac", "brew install typescript")
  ]
];
const whatTheShell = {
  name: "what-the-shell",
  description: "Use GitHub Copilot to get shell commands from natural language descriptions.",
  promptFn: fewShotQueryPrompter(shellPromptHeadline, shellFewshotExamples),
  invariant: (completion) => completion.trim() !== DONTKNOW
};

const gitPromptHeadline = assistHeadline("git commands");
const gitFewshotExamples = [
  [
    interact("switch branch", "git switch <branch>"),
    interact("main", "git switch main")
  ],
  [
    interact("Delete a branch", "git branch -d <branch>"),
    interact("branch new-feature", "git branch -d new-feature")
  ],
  [interact("Switch to the previous branch", "git switch -")],
  [
    interact(
      "find commit messages with the word security",
      `git log --grep="security" --oneline`
    ),
    interact(
      "by user johnDoe177",
      `git log --grep="security" --author="johnDoe177" --oneline`
    )
  ],
  [
    interact("undo changes", "git restore <file or tree>"),
    interact("to before I added the new feature", DONTKNOW),
    interact(
      "to before the last change to src/engine.ts",
      [
        "COMMIT_HASH=$(git log --oneline src/engine.ts | tail -n 1 | cut -d ' ' -f 1)",
        "git checkout $COMMIT_HASH"
      ].join("\n")
    )
  ],
  [
    interact("find security fixes", DONTKNOW),
    interact(
      "commits with security in the message",
      `git log --grep="security" --oneline`
    ),
    interact(
      "csv formatted as sha,author,date",
      `git log --grep="security" --pretty=format:"%h,%an,%ad" --date=short`
    )
  ],
  [
    interact(
      "What changes were made by taylorotwell?",
      `git log -p --author=taylorotwell *`
    )
  ],
  [
    interact(
      'Move the latest commit to new branch "new-feature"',
      [
        "git branch new-feature",
        "git reset --hard HEAD~1",
        "git switch new-feature"
      ].join("\n")
    ),
    interact(
      "Move the last two commits",
      [
        "git branch new-feature",
        "git reset --hard HEAD~2",
        "git switch new-feature"
      ].join("\n")
    )
  ],
  [
    interact("untangle my mess", DONTKNOW),
    interact("find the commit that broke the build", `git bisect start`)
  ],
  [interact("push and overwrite", "git push --force-with-lease")],
  [interact("Throw away all my local changes", "git restore .")],
  [
    interact('rebase onto "prev-feature"', "git rebase prev-feature"),
    interact(
      "previously rebased before commit a2b3c4",
      "git rebase --onto prev-feature a2b3c4"
    )
  ],
  [
    interact(
      "Commits made by KaiserSoze404 if the commit touches src/index.ts",
      `git log -p --author=KaiserSoze404 --oneline -- src/index.ts`
    )
  ]
];
const gitAssist = {
  name: "git-assist",
  description: "Translate a natural language description of a git command to an actual git command.",
  promptFn: fewShotQueryPrompter(gitPromptHeadline, gitFewshotExamples),
  invariant: (completion) => completion.trim() !== ""
};

const ghPromptHeadline = assistHeadline("GitHub CLI commands");
const ghFewshotExamples = [
  [
    interact("list issues", "gh issue list"),
    interact("all", "gh issue list -L 10000")
  ],
  [
    interact(
      "issues for tailwindlabs/tailwindcss",
      "gh issue list -R tailwindlabs/tailwindcss"
    )
  ],
  [
    interact(
      "users who have created issues as json",
      `gh issue list --state="all" -L 10000 --json author`
    )
  ],
  [
    interact(
      "default branch on jquense/yup",
      `gh api repos/jquense/yup | jq .default_branch`
    )
  ],
  [
    interact(
      "the message of the most recent commit on jquense/yup",
      `gh api repos/jquense/yup/commits --jq '. | sort_by(.commit.author.date) | last | .commit.message'`
    )
  ],
  [
    interact(
      "the message of the most recent commit",
      `CUR_REPO_NWO=$(gh repo view --json nameWithOwner --jq '.nameWithOwner') gh api repos/$CUR_REPO_NWO/commits --jq '. | sort_by(.commit.author.date) | last | .commit.message'`
    )
  ],
  [
    interact(
      "pull requests in microsoft/vscode",
      `gh pr list -R microsoft/vscode`
    ),
    interact(
      "title and status, sorted by title",
      `gh pr list -R microsoft/vscode --json title,state --jq '. | sort_by(.title)'`
    ),
    interact(
      "printed human readable",
      `gh pr list -R microsoft/vscode --json title,state --jq '. | sort_by(.title) | map(.title + " " + .state) | .[]'`
    )
  ],
  [
    interact(
      "title and last modified date of all pull requests in tailwindlabs/tailwind-css as json",
      `gh pr list -R tailwindlabs/tailwind-css -L 10000 --json title,updatedAt`
    )
  ],
  [
    interact(
      "how many issues are there on tailwindlabs/tailwind-css",
      `gh issue list -R tailwindlabs/tailwind-css -L 10000 --json number --jq ". | length"`
    )
  ],
  [
    interact("closed issues for this repo", `gh issue list --state="closed"`),
    interact(
      "in the last week",
      `LASTWEEK=$(date -I -d "7 days ago") gh issue list -L 10000 --state="closed" --json number,title,closedAt --jq '.[] | select(.closedAt > "$LASTWEEK")`
    ),
    interact(
      "sorted",
      `LASTWEEK=$(date -I -d "7 days ago") gh issue list -L 10000 --state="closed" --json closedAt,number,title --jq '.[] | select(.closedAt > "$LASTWEEK")' | sort`
    )
  ]
];
const ghAssist = {
  name: "gh-assist",
  description: "Convert plain english to GitHub CLI commands.",
  promptFn: fewShotQueryPrompter(ghPromptHeadline, ghFewshotExamples),
  invariant: (completion) => completion.trim() !== ""
};

async function verifyWaitlistAccess(accessToken) {
  var _a;
  const res = await axios.get(
    NEXT_WAITLIST_URL + `/api/verify`,
    {
      params: {
        project: WAITLIST_PROJECT_ID,
        token: accessToken
      }
    }
  );
  let hasAccess = (_a = res.data) == null ? void 0 : _a.hasAccess;
  if (!hasAccess) {
    throw new Error(
      `You do not have access to GitHub Copilot CLI. Please join the waitlist at https://githubnext.com/projects/copilot-cli/.`
    );
  }
}
async function getCopilotToken(accessToken) {
  const res = await axios.get(COPILOT_TOKEN_URL, {
    headers: {
      Authorization: `token ${accessToken}`
    }
  });
  return res.data.token;
}
function makeCopilotProxy() {
  let instance = axios.create({
    baseURL: `https://copilot-proxy.githubusercontent.com/v1/engines/copilot-labs-codex`
  });
  instance.interceptors.request.use(
    (config) => {
      let token = getCachedCopilotToken();
      config.headers = {
        Authorization: `Bearer ${token}`,
        "X-Request-Start": `t=${Date.now()}`
      };
      return config;
    },
    (error) => {
      Promise.reject(error);
    }
  );
  instance.interceptors.response.use(
    function(response) {
      return response;
    },
    async function(error) {
      var _a;
      if (!axios.isAxiosError(error)) {
        return Promise.reject(error);
      }
      const originalRequest = error.config;
      let cachedCopilotToken = getCachedCopilotToken();
      let cachedAccessToken = getCachedAccessToken();
      if (((_a = error == null ? void 0 : error.response) == null ? void 0 : _a.status) === 401 && cachedCopilotToken && cachedAccessToken && // @ts-ignore
      !originalRequest._retry) {
        originalRequest._retry = true;
        let copilotToken = await getCopilotToken(cachedAccessToken);
        setCachedCopilotToken(copilotToken);
        return instance(originalRequest);
      }
      return Promise.reject(error);
    }
  );
  return {
    getCompletion: (params, config) => {
      return instance.post(
        `/completions`,
        {
          ...params,
          stream: true
        },
        config
      );
    }
  };
}

const auth = new Command("auth");
async function triggerKeypress() {
  process.stdin.setRawMode(true);
  return new Promise(
    (resolve) => process.stdin.once("data", (data) => {
      const byteArray = [...data];
      if (byteArray.length > 0 && byteArray[0] === 3) {
        console.log("^C");
        process.exit(1);
      }
      process.stdin.setRawMode(false);
      resolve();
    })
  );
}
auth.description("Retrieve and store a GitHub Access Token.").action(async function() {
  var _a;
  let spinner = ora();
  const userCodeResponse = await axios.post(
    DEVICE_CODE_URL,
    {
      client_id: CLIENT_ID,
      scope: "read:user"
    },
    {
      headers: { Accept: "application/json" }
    }
  );
  console.log(
    `Copy this code: ${chalk.bold.underline(
      userCodeResponse.data.user_code
    )}
`
  );
  console.log(
    `Then ${chalk.bold(
      "press any key to launch the authorization page"
    )}, paste the code in and approve the access.`
  );
  spinner.start(
    `It will take up to ${userCodeResponse.data.interval} seconds after approval for the token to be retrieved. 
`
  );
  await triggerKeypress();
  open(userCodeResponse.data.verification_uri);
  let expiresIn = userCodeResponse.data.expires_in;
  let accessToken;
  while (expiresIn > 0) {
    const accessTokenResponse = await axios.post(
      ACCESS_TOKEN_URL,
      {
        client_id: CLIENT_ID,
        device_code: userCodeResponse.data.device_code,
        grant_type: "urn:ietf:params:oauth:grant-type:device_code"
      },
      {
        headers: { Accept: "application/json" }
      }
    );
    expiresIn -= userCodeResponse.data.interval;
    await new Promise(
      (resolve) => setTimeout(resolve, 1e3 * userCodeResponse.data.interval)
    );
    if (accessTokenResponse.data["access_token"]) {
      spinner.stop();
      accessToken = accessTokenResponse.data["access_token"];
      break;
    }
  }
  if (accessToken === void 0) {
    spinner.stop();
    console.log("Timed out waiting for authorization");
    throw new Error("Timed out waiting for authorization");
  } else {
    spinner.stop();
    setCachedAccessToken(accessToken);
    try {
      await verifyWaitlistAccess(accessToken);
    } catch (e) {
      let message = (_a = e.message) != null ? _a : "Unable to verify waitlist access.";
      console.log(`Failed to authenticate: ${chalk.bold.red(message)}`);
      process.exit(1);
    }
    try {
      let copilotToken = await getCopilotToken(accessToken);
      setCachedCopilotToken(copilotToken);
      process.exit(0);
    } catch (e) {
      console.log(
        `
Failed to authenticate: ${chalk.bold.red(
          "GitHub Copilot access is required to use this CLI. Please sign up at https://github.com/features/copilot."
        )}
`
      );
      process.exit(1);
    }
  }
});

const alias = new Command("alias");
const modeAliases = {
  "what-the-shell": "??",
  "git-assist": "git?",
  "gh-assist": "gh?"
};
function injectHistory(command_var, env) {
  if (env.shellName === "bash") {
    return `history -s $(history 1 | cut -d' ' -f4-); history -s "$${command_var}";`;
  } else if (env.shellName === "zsh") {
    return `print -s "$${command_var}";`;
  }
  return "";
}
function buildFunction(shellFn, command, env) {
  return `
  ${shellFn} () {
    TMPFILE=$(mktemp);
    trap 'rm -f $TMPFILE' EXIT;
    if ${command} --shellout $TMPFILE; then
      if [ -e "$TMPFILE" ]; then
        FIXED_CMD=$(cat $TMPFILE);
        ${injectHistory("FIXED_CMD", env)}
        eval "$FIXED_CMD"
      else
        echo "Apologies! Extracting command failed"
      fi
    else
      return 1
    fi
  };`;
}
function buildAlias(alias2, shellFn) {
  return `alias '${alias2}'='${shellFn}';`;
}
function extractShellName(shellPath) {
  let shellName = shellPath.split("/").pop();
  if (shellName == null ? void 0 : shellName.startsWith("-")) {
    shellName = shellName.slice(1);
  }
  return shellName != null ? shellName : "unknown";
}
alias.summary("Add convenient GitHub Copilot CLI aliases to your shell.").description(
  dedent`
    Add convenient GitHub Copilot CLI aliases to your shell.

    Run as

      $ eval "$(github-copilot-cli alias -- "$0")"

    This will add three commands to your shell:
    
    - \`??\` <ask about any type of task in the shell>
    - git? <ask about git tasks>
    - gh? <ask about GitHub CLI tasks>

    You can put this invocation in your .bashrc or .zshrc to add the aliases to
    all shell sessions. See the README for installation instructions.
    `
).argument(
  "<shell path>",
  'The path of your shell. Put "$0" here to denote your current shell.'
).action((shellPath) => {
  const env = {
    shellName: extractShellName(shellPath)
  };
  let currentProcess = process.argv[1];
  function runWithArg(cmd) {
    return `${currentProcess} ${cmd} "$@"`;
  }
  let aliases = [];
  for (const [mode, alias2] of Object.entries(modeAliases)) {
    const shellFn = `copilot_${mode}`;
    aliases.push(buildFunction(shellFn, runWithArg(mode), env));
    aliases.push(buildAlias(alias2, shellFn));
  }
  aliases = [
    ...aliases,
    buildAlias("wts", "copilot_what-the-shell")
    // second alias
  ];
  console.log(aliases.join("\n"));
});

function tokenPresenceHook(thisCommand) {
  try {
    tokenPresenceCheck();
  } catch (e) {
    thisCommand.error(
      "Authentication error. Please run `github-copilot-cli auth` to re-authenticate."
    );
  }
}

function QueryHistory() {
  let history = useHistory();
  let showQueryHistory = history.interactions.length > 1;
  if (!showQueryHistory)
    return null;
  return /* @__PURE__ */ jsx(Box, { flexDirection: "column", children: /* @__PURE__ */ jsx(Box, { paddingY: 1, children: history.interactions.map((interaction, idx) => /* @__PURE__ */ jsxs(
    Text,
    {
      dimColor: Boolean(interaction.response.type === "refuse"),
      children: [
        idx + 1,
        ") ",
        interaction.query,
        " "
      ]
    },
    idx
  )) }) });
}

function Spinner({ children, type = "dots" }) {
  return /* @__PURE__ */ jsxs(Box, { children: [
    /* @__PURE__ */ jsx(Text, { color: "red", children: "Loading..." }),
    /* @__PURE__ */ jsx(Box, { marginLeft: 0, children: /* @__PURE__ */ jsx(Text, { children }) })
  ] });
}

async function parseStreamResponse(response) {
  let data = await getStream(response.data, {
    encoding: "utf-8"
  });
  if (data.trim() === `data: [DONE]`) {
    throw new Error("No completion available.");
  }
  if (!data.includes("[DONE]")) {
    throw new Error("Stream did not complete.");
  }
  data = data.replace(`data: [DONE]`, "");
  data = data.replace(/data: /, "");
  let dataArray = data.split("data: ");
  dataArray = dataArray.map((d) => d.replace(/\n/g, ""));
  let parsedItems = dataArray.map((d) => JSON.parse(d));
  return parsedItems.map((d) => {
    return d.choices[0].text;
  }).join("");
}

function useCompletion({
  history,
  options,
  invariant,
  promptFn
}) {
  const { handleResponse } = useActions();
  let enabled = history.lastQuery !== void 0;
  return useQuery(
    ["completion", iteration(history)],
    async () => {
      const prompt = promptFn(history);
      let httpResponse = await proxy.getCompletion({
        ...options,
        prompt
      });
      let parsedResponse = await parseStreamResponse(httpResponse);
      let answer = answerFromCompletion(parsedResponse);
      let response;
      if (!invariant(answer)) {
        response = {
          type: "refuse",
          message: "Sorry. GitHub Copilot doesn't know how to do that."
        };
      } else {
        response = { type: "completion", completion: answer };
      }
      return response;
    },
    {
      enabled,
      onSettled: (modelResponse) => {
        handleResponse(modelResponse);
        trackModelResponse();
      }
    }
  );
}
const explanationPromptHeader = dedent`Structured and complete explanations of shell commands.
  We follow the commands piece-by-piece in bullet-point form, with multiple statements being separate bullets.
  Sub-commands form sub-bullets.
  All flags are explained as sub-bullets.
  `;
const explanationFewshotExamples = [
  explanation(`find -name "*.ts" | grep -v node_modules`, [
    {
      hd: `\`find\` is used to list files.`,
      sub: [
        `\`-name "*.ts"\` stipulates that we search for files ending in \`.ts\`.`
      ]
    },
    {
      hd: `\`| grep\` means we pass that list of files to \`grep\` which filters the list.`,
      sub: [
        `\`-v node_modules\` means we exclude all files containing node_modules in its path.`
      ]
    }
  ]),
  explanation(`sudo apt-get install typescript`, [
    {
      hd: `\`sudo\` is used to run a command with elevated rights, to e.g. change system files.`,
      sub: [
        {
          hd: `\`apt-get\` is the Ubuntu package manager.`,
          sub: [
            `\`install typescript\` directs it to install the \`typescript\` package.`
          ]
        }
      ]
    }
  ]),
  explanation(
    dedent`for file in $(find -name "*.py"); do
      echo "$file";
      wc -l $file;
    done`,
    [
      {
        hd: `The for loop iterates over a list of items and executes its body for each, using the loop variable \`$file\`.`,
        sub: [
          {
            hd: `The list of items is the output of a sub-call to \`find\` which is used to list files.`,
            sub: [
              `\`-name "*.py"\` stipulates that we search for files ending in \`.py\`.`
            ]
          }
        ]
      },
      {
        hd: `The loop body executes two commands for each file:`,
        sub: [
          `\`echo "$file"\` will print the name of the current file to the terminal.`,
          {
            hd: `\`wc\` is used to count the number of lines in the current file.`,
            sub: [
              `\`-l\` specifies that we only want the line count, not the word or character count.`
            ]
          }
        ]
      }
    ]
  ),
  explanation(
    dedent`
    git branch new-feature
    git reset --hard HEAD~1
    git switch new-feature
    `,
    [
      `\`git branch new-feature\` creates a new branch called \`new-feature\`.`,
      {
        hd: `\`git reset\` resets the current branch to a previous commit.`,
        sub: [
          `\`--hard\` means that we also discard any changes made to the files in the working directory.`,
          `\`HEAD~1\` specifies that we reset to the commit one before the current one.`
        ]
      },
      `\`git switch new-feature\` switches to the \`new-feature\` branch.`
    ]
  ),
  explanation(`T=$(grep TITLE $folder/Makefile | cut -d "=" -f 2);`, [
    {
      hd: `\`T=$(grep TITLE $folder/Makefile | cut -d "=" -f 2)\` sets the variable \`T\` to the title of the folder.`,
      sub: [
        {
          hd: `\`T=$(...)\` is used to assign the output of the command inside \`$(...)\` to the variable \`T\`.`,
          sub: [
            `\`grep TITLE $folder/Makefile\` searches for the line containing \`TITLE\` in the \`Makefile\` of the current folder.`,
            `| \`cut -d "=" -f 2\` takes this line, splits it at the \`=\` character and returns the second part.`
          ]
        }
      ]
    }
  ]),
  explanation(`ls -lart`, [
    {
      hd: `\`ls\` is used to list files.`,
      sub: [
        `\`-l\` specifies that we want a long listing.`,
        `\`-a\` specifies that we want to include hidden files.`,
        `\`-r\` specifies that we want to reverse the order of the listing.`,
        `\`-t\` specifies that we want to sort the listing by modification time.`
      ]
    }
  ]),
  explanation(`curl ifconfig.me`, [
    {
      hd: `\`curl\` is used to issue web requests, e.g. download web pages.`,
      sub: [
        `\`ifconfig.me\` is a web service that returns the public IP address of the requestor.`
      ]
    }
  ])
];
const explanationPrompter = fewShotPrompter(
  explanationPromptHeader,
  explanationFewshotExamples,
  (cmd) => formatQA(cmd.trim(), void 0, "question")
);
function useExplanation(command, config) {
  return useQuery(
    ["explanation", command],
    async () => {
      const prompt = explanationPrompter(command);
      let response = await proxy.getCompletion({
        max_tokens: 1024,
        n: 1,
        logprobs: void 0,
        temperature: 0,
        top_p: 1,
        stop: ["---"],
        prompt
      });
      const explanation2 = answerFromCompletion(
        await parseStreamResponse(response)
      );
      return explanation2;
    },
    config
  );
}

function Completion({ mode }) {
  const history = useHistory();
  const { status, data: modelResponse } = useCompletion({
    history,
    promptFn: mode.promptFn,
    options: {
      n: 1,
      stop: ["---", "\nR:\n"],
      // stop at end of interaction, or at next round if model expects that
      max_tokens: 256,
      logprobs: void 0,
      temperature: 0,
      top_p: 1,
      ...mode.promptOptions
    },
    invariant: mode.invariant
  });
  return /* @__PURE__ */ jsxs(Box, { paddingY: 1, children: [
    status === "loading" && /* @__PURE__ */ jsx(Spinner, { type: "timeTravel", children: "Hold on, asking GitHub Copilot..." }),
    status === "success" && modelResponse.type === "completion" && /* @__PURE__ */ jsx(Text, { children: highlight(modelResponse.completion.trimEnd(), {
      language: "bash"
    }) }),
    status === "success" && modelResponse.type === "refuse" && /* @__PURE__ */ jsxs(Text, { color: "yellow", children: [
      "\u26A0\uFE0F ",
      modelResponse.message
    ] }),
    status === "error" && /* @__PURE__ */ jsx(Text, { color: "red", bold: true, children: "Error fetching completion." })
  ] });
}

const successItems = [
  {
    label: "\u2705 This looks right, thank you!",
    value: "correct"
  },
  {
    label: "\u{1F914} Actually, I can be more specific. Let me clarify!",
    value: "revise"
  },
  {
    label: "\u274C Cancel",
    value: "cancel"
  }
];
const errorItems = [
  {
    label: "\u{1F914} Actually, I can be more specific. Let me clarify!",
    value: "revise"
  },
  {
    label: "\u274C Cancel",
    value: "cancel"
  }
];
function CompletionChoices() {
  const { setState, setExitReason } = useActions();
  const isError = isRefuseResponse(useHistory());
  const { exit } = useApp();
  const items = isError ? errorItems : successItems;
  const handleSelect = async (item) => {
    switch (item.value) {
      case "cancel":
        setExitReason("cancel");
        await trackExit("cancel");
        exit();
      case "correct":
        setState("confirm");
        break;
      case "revise":
        setState("revise");
        break;
    }
  };
  return /* @__PURE__ */ jsx(SelectInput, { items, onSelect: handleSelect });
}

function ConfirmCompletion({ shellOut }) {
  const { setState } = useActions();
  const lastResult = useLastCompletion();
  const { exit } = useApp();
  const { mutateAsync, status } = useMutation(trackConfirmCommandRun);
  useInput(async (input, key) => {
    if (status !== "idle")
      return;
    if (key.escape) {
      setState("response");
    } else if (input === "y" || input === "Y") {
      await mutateAsync();
      if (shellOut)
        fs.writeFileSync(shellOut, lastResult);
      exit();
    } else if (input === "n" || input === "N") {
      trackConfirmCommandCancel();
      setState("response");
    }
  });
  useEffect(() => {
    if (shellOut === void 0 || shellOut === "") {
      exit();
    }
  }, [shellOut]);
  if (status === "loading")
    return /* @__PURE__ */ jsx(Spinner, { type: "timeTravel", children: "Hold on, executing commmand..." });
  if (shellOut === void 0 || shellOut === "") {
    return /* @__PURE__ */ jsxs(Box, { flexDirection: "column", children: [
      /* @__PURE__ */ jsx(Text, { children: "Great! Just copy/paste the command above into your shell." }),
      /* @__PURE__ */ jsx(Box, { marginTop: 1, marginBottom: 1, children: /* @__PURE__ */ jsx(Text, { children: "(If you want accepted commands to be run without manual copy/paste, consider installing the `alias` commands. See https://www.npmjs.com/package/@githubnext/github-copilot-cli#alias-convenience-commands for instructions.)" }) })
    ] });
  } else {
    return /* @__PURE__ */ jsxs(Box, { marginLeft: 2, marginTop: 1, flexDirection: "column", children: [
      /* @__PURE__ */ jsx(Box, { children: /* @__PURE__ */ jsx(Text, { children: "This will execute the suggested command in your shell." }) }),
      /* @__PURE__ */ jsx(Box, { flexDirection: "row", children: /* @__PURE__ */ jsx(Box, { marginRight: 1, marginBottom: 1, children: /* @__PURE__ */ jsx(Text, { children: "Are you sure? (y/N)" }) }) })
    ] });
  }
}

marked.setOptions({
  // Define custom renderer
  renderer: new TerminalRenderer()
});
const Placeholder = () => /* @__PURE__ */ jsx(Text, { dimColor: true, children: "Awaiting command..." });
const LIST_SYMBOLS = ["\u25CB", "\u25C6", "\u25C7", "\u25AA", "\u25AB", "\u2014"];
function markupExplanation(explanation) {
  function formatLine(line) {
    if (!line)
      return "";
    const [beforeStar, afterStar] = line.split("* ");
    const level = beforeStar.length / EXPLAIN_LIST_INDENT.length;
    return `${beforeStar}${LIST_SYMBOLS[level % LIST_SYMBOLS.length]} ${marked(
      afterStar
    ).trim()}`;
  }
  return explanation.split("\n").map((line, i) => {
    return /* @__PURE__ */ jsx(Box, { margin: 0, children: /* @__PURE__ */ jsx(Text, { children: formatLine(line) }) }, i);
  });
}
function Explanation() {
  const completion = useLastCompletion();
  const queryClient = useQueryClient();
  const { data: rawExplanation, status } = useExplanation(completion, {
    enabled: Boolean(completion),
    onSuccess: (explanation2) => {
      trackExplanationSuccess(explanation2);
    },
    onError: (e) => {
      trackExplanationError(e.message);
    }
  });
  let explanation = dedent("\n" + rawExplanation || "");
  const isFetchingCompletion = Boolean(queryClient.isFetching(["completion"]));
  if (isFetchingCompletion) {
    return /* @__PURE__ */ jsx(Box, { paddingY: 1, children: /* @__PURE__ */ jsx(Placeholder, {}) });
  }
  return /* @__PURE__ */ jsxs(Box, { paddingY: 1, children: [
    status === "idle" && /* @__PURE__ */ jsx(Placeholder, {}),
    status === "loading" && /* @__PURE__ */ jsx(Spinner, { type: "timeTravel", children: "Hold on, asking GitHub Copilot..." }),
    status === "success" && /* @__PURE__ */ jsx(Box, { display: "flex", flexDirection: "column", margin: 0, children: markupExplanation(explanation) }),
    status === "error" && /* @__PURE__ */ jsx(Text, { color: "red", bold: true, children: "Error fetching explanation." })
  ] });
}

function RevisePrompt() {
  const { handleRevision, setState } = useActions();
  const queryClient = useQueryClient();
  useInput((_, key) => {
    if (key.escape) {
      setTimeout(() => {
        setState("response");
      });
    }
  });
  const onSubmit = async (revision) => {
    if (!revision) {
      setState("response");
      return;
    }
    handleRevision(revision);
    trackRevisionRequest();
    await queryClient.refetchQueries(["completion"]);
  };
  return /* @__PURE__ */ jsx(Fragment, { children: /* @__PURE__ */ jsx(Box, { paddingY: 1, children: /* @__PURE__ */ jsxs(Box, { display: "flex", alignItems: "center", children: [
    /* @__PURE__ */ jsx(Box, { marginRight: 1, children: /* @__PURE__ */ jsx(Text, { children: "Enter your revision:" }) }),
    /* @__PURE__ */ jsx(UncontrolledTextInput, { onSubmit })
  ] }) }) });
}

const proxy = makeCopilotProxy();
function ReviseLoopInner(props) {
  const state = useState();
  const history = useHistory();
  const isRefuse = isRefuseResponse(history);
  return /* @__PURE__ */ jsxs(Box, { marginTop: 1, flexDirection: "column", children: [
    /* @__PURE__ */ jsx(QueryHistory, {}),
    /* @__PURE__ */ jsx(Completion, { ...props }),
    !isRefuse && /* @__PURE__ */ jsx(Fragment, { children: /* @__PURE__ */ jsx(Explanation, {}) }),
    state === "response" && /* @__PURE__ */ jsx(CompletionChoices, {}),
    state === "confirm" && /* @__PURE__ */ jsx(ConfirmCompletion, { ...props }),
    state === "revise" && /* @__PURE__ */ jsx(RevisePrompt, {})
  ] });
}
let SIG_INPUTs = ["c", "d", "\\"];
function ReviseLoop(props) {
  const app = useApp();
  const { setExitReason } = useActions();
  const client = new QueryClient({
    defaultOptions: {
      queries: {
        retry: false,
        cacheTime: 0,
        refetchOnWindowFocus: false,
        refetchOnReconnect: false
      }
    }
  });
  async function handleExit() {
    setExitReason("system");
    await trackExit("system");
    app.exit();
  }
  useInput(async (input, key) => {
    let exitConditions = [
      input.charCodeAt(0) === 28,
      key.ctrl && SIG_INPUTs.includes(input)
    ];
    if (exitConditions.some((c) => c)) {
      await handleExit();
      return;
    }
  });
  return /* @__PURE__ */ jsx(QueryClientProvider, { client, children: /* @__PURE__ */ jsx(ReviseLoopInner, { ...props }) });
}

async function runReviseLoop(initialInput, mode, shellOut) {
  useCliStore.setState(makeInitialState(initialInput, mode.name));
  trackInvocation();
  const app = render(
    /* @__PURE__ */ jsx(ReviseLoop, { initialInput, mode, shellOut }),
    { exitOnCtrlC: false }
  );
  return app.waitUntilExit();
}

const version = version$1;
const program = new Command();
program.name("Copilot CLI").description(
  "A CLI experience for letting GitHub Copilot help you on the command line."
).version(version);
program.addCommand(auth);
program.addCommand(alias);
for (const mode of [whatTheShell, gitAssist, ghAssist]) {
  const prg = new Command(mode.name);
  prg.description(mode.description).option("--shellout <shell-out-file>", "Used by the shell alias.").argument("<query...>", "Description of the thing you want to do.").hook("preAction", tokenPresenceHook).hook("preAction", initializeTelemetry).hook("postAction", async () => {
    const { exitReason } = useCliStore.getState();
    await flushTelemetry();
    process.exit(exitReason == "system" || exitReason == "cancel" ? 1 : 0);
  }).action(async function() {
    const query = this.args.join(" ");
    await runReviseLoop(query, mode, this.opts().shellout);
  });
  program.addCommand(prg);
}
program.parseAsync();

export { version };
